/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2017/18      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : lecture 3
///
///  author     : lessig@isg.cs.ovgu.de
///
///  project    : GPU Programming
///
///  description: race conditions
///
////////////////////////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iostream>
#include <iomanip>
#include <vector>
#include <thread>
#include <cstdlib>
#include <chrono>
#include <ctime>
#include <cassert>
#include <algorithm>

int* scratchref = nullptr;

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Binary search
//! returns pointer to largest element smaller than token if the element does not exist
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
T*
binary_search( T* first, T* last, const T& token) {

  int* mid = first + (last - first) / 2;

  while( (last - first) > 1) {

    if( *mid == token) {
      break;
    }

    if( *mid > token) {
      last = mid;
      mid = first + (last - first) / 2;
    }
    else {
      first = mid;
      mid = first + (last - first) / 2;
    }
  }

  return mid;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge step of merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
merge( T* data, T* last, int nhalf, T* scratch) {

  int* data1 = data;
  int* last1 = data + nhalf;
  int* data2 = data + nhalf;
  int* last2 = last;

  // merge
  T* scratchref = scratch;
  while( data1 != last1 && data2 != last2) {
    T*& which = (*data1 < *data2) ? data1 : data2;
    *scratch++ = std::move( *which++);
  }
  // copy remainder (obviously, only one will have an effect)
  std::copy( data1, last1, scratch);
  std::copy( data2, last2, scratch);

  // put sorted items into data
  std::copy( scratchref, scratchref + (last-data), data);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge step of merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
merge( T* data1, T* last1, T* data2, T* last2, T* scratch) {

  T* data1ref = data1;
  int length = last2 - data1;

  // merge
  T* scratchref = scratch;
  while( data1 != last1 && data2 != last2) {
    T*& which = (*data1 < *data2) ? data1 : data2;
    *scratch++ = std::move( *which++);
  }
  // copy remainder (obviously, only one will have an effect)
  std::move( data1, last1, scratch);
  std::move( data2, last2, scratch);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge step of merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
mergeParallel( T* data1, T* last1, T* data2, T* last2, T* scratch, int num_threads) {

  const int nmin = 8;

  int length1 = last1 - data1;
  int length2 = last2 - data2;

  if( length1 + length2 <= nmin) {
    merge( data1, last1, data2, last2, scratch);
    return;
  }

  // split longer list
  T* h1 = nullptr;
  T* h2 = nullptr;
  if( length1 < length2) {
    h2 = data2 + length2 / 2;
    h1 = std::upper_bound( data1, last1, *h2);
  }
  else {
    h1 = data1 + length1 / 2;
    h2 = std::lower_bound( data2, last2, *h1);
  }

  // split up scratch
  T* scratch2 = scratch + (h1 - data1) + (h2 - data2);

  // divide step
  if( num_threads > 0) {
    // std::cerr << "Launching thread: " << num_threads << std::endl;
    // start just one new thread and use the current one to do half the work
    std::thread th1( mergeParallel<T>, data1, h1, data2, h2, scratch, num_threads-2);
    std::thread th2( mergeParallel<T>, h1, last1, h2, last2, scratch2, num_threads-2);
    th1.join();
    th2.join();
  }
  else {
    merge( data1, h1, data2, h2, scratch);
    merge( h1, last1, h2, last2, scratch2);
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
mergesort( T* data, T* last, T* scratch) {

  const int nmin = 2;
  int nhalf = (last - data) / 2;

  if( nhalf < nmin) {
    std::sort( data, last);
    return;
  }

  // recursion
  mergesort( data, data + nhalf, scratch);
  mergesort( data + nhalf, last, scratch + nhalf);

  // merge step
  merge( data, last, nhalf, scratch);

}

////////////////////////////////////////////////////////////////////////////////////////////////////
//! Merge sort
////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename T>
void
mergesortParallel( T* data, T* last, T* scratch, const int num_threads) {

  int chunk_size = (last - data) / num_threads;
  assert( 0 == (last - data) % num_threads);
  assert( 0 == num_threads % 2);

  // run merge sort on chunks
  std::vector< std::thread > threads;
  for( int tid = 0; tid < num_threads; ++tid) {
    threads.push_back( std::thread( mergesort<T>, data + tid*chunk_size,
                                                  data + (tid+1)*chunk_size,
                                                  scratch + tid*chunk_size));
  }

  for( auto& th : threads) {
    th.join();
  }

  // merge chunks, traversing the tree upwards towards the root
  int active_chunks = num_threads;
  while( active_chunks > 1) {

#if 0
    // merge adjacent nodes in the tree
    for( int i = 0; i < active_chunks; i+=2) {
      merge( data + i * chunk_size, data + (i+2) * chunk_size, chunk_size,
             scratch + i * chunk_size);
    }

#else

    assert( 0 == num_threads % active_chunks);

    std::vector< std::thread > threads;
    // merge adjacent nodes in the tree
    for( int i = 0; i < active_chunks; i+=2) {

      threads.push_back( std::thread( mergeParallel<T>,
                                      data + i * chunk_size, data + (i+1) * chunk_size,
                                      data + (i+1) * chunk_size, data + (i+2) * chunk_size,
                                      scratch + i * chunk_size, num_threads / chunk_size-1));
    }

    for( int i = 0; i < active_chunks; i+=2) {
      threads[i/2].join();
      std::move( scratch + i * chunk_size, scratch + (i+2) * chunk_size, data + i * chunk_size);
    }
#endif

    chunk_size *= 2;
    active_chunks /= 2;
  }
}


////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  std::srand( std::time(0));

  // allocate memory
  int n = 8192 * 8192;
  int* data = (int*) malloc( n * sizeof(int));
  int* scratch = (int*) malloc( n * sizeof(int));
  scratchref = scratch;
  std::generate( data, data + n, std::rand);

  // reference solution
  int* dataref = (int*) malloc( n * sizeof(int));
  memcpy( dataref, data, n * sizeof(int));
  std::sort( dataref, dataref+n);

#if 0
  // run merge sort
  mergesort( data, data+n, scratch);
#endif

#if 1
  // run parallel merge sort
  unsigned int num_threads = std::thread::hardware_concurrency();
  std::cout << "num_threads = " << num_threads << std::endl;

  // start timing
  clock_t tstart = clock();

  for( int rep = 0; rep < 10; ++rep) {
    mergesortParallel( data, data+n, scratch, num_threads);
  }

  clock_t tend = clock();

  // final time
  double telapsed = (double)(tend - tstart) / CLOCKS_PER_SEC;
  printf( "time elaspsed = %2.4f sec\n", telapsed);
#endif

  // check correctness of result
  bool correct = true;
  for( int i = 0; i < n; ++i) {
    correct &= (dataref[i] == data[i]);
  }
  std::cout << "Mergesort " << ((correct) ? "succeeded." : "failed.") << std::endl;

  // clean up
  free( data);
  free( scratch);
  free( dataref);

  return EXIT_SUCCESS;
}
